 /**************************************************************************
  *                                                                        *
  *               Copyright (C) 1996, Silicon Graphics, Inc.               *
  *                          All Rights Reserved.                          *
  *                                                                        *
  *  The files in this subtree  contain  UNPUBLISHED  PROPRIETARY  SOURCE  *
  *  CODE of Silicon Graphics, Inc.;  the  contents  of  these  files may  *
  *  not be disclosed to third  parties,  copied  or  duplicated  in  any  *
  *  form, in whole or in part,  without  the  prior  written  permission  * 
  *  of  Silicon Graphics, Inc.                                            *
  *                                                                        *
  *  RESTRICTED RIGHTS LEGEND:                                             *
  *  Use,  duplication  or  disclosure  by  the  Government is subject to  *
  *  restrictions as set forth in subdivision (c)(1)(ii) of the Rights in  *
  *  Technical Data and Computer Software clause at  DFARS  252.227-7013,  *
  *  and/or in similar or successor  clauses in the FAR,  DOD or NASA FAR  *
  *  Supplement.  Unpublished - rights reserved  under the Copyright Laws  *
  *  of the United States.                                                 *
  *                                                                        *
  *  THIS SOFTWARE IS  PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,  *
  *  EXPRESS,  IMPLIED OR  OTHERWISE,  INCLUDING  WITHOUT LIMITATION, ANY  *
  *  WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.      *
  *                                                                        *
  *  IN  NO  EVENT  SHALL  SILICON  GRAPHICS  BE  LIABLE FOR ANY SPECIAL,  *
  *  INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF  ANY  KIND,  OR ANY  *
  *  DAMAGES WHATSOEVER RESULTING FROM LOSS  OF  USE,  DATA  OR  PROFITS,  *
  *  WHETHER OR NOT ADVISED OF  THE  POSSIBILITY  OF  DAMAGE,  AND ON ANY  *
  *  THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR  *
  *  PERFORMANCE OF THIS SOFTWARE.                                         *
  **************************************************************************/

/*
    Author : Patrick Bouchaud
             galaad@neu.sgi.com
*/

#include <stdlib.h>
#include <stdio.h>
#include <GL/gl.h>
#include <math.h>
#include "data.h"

#include "culling.h"

/*---------------------------------------------------------------------------*/
/*			XCullDraw
/*---------------------------------------------------------------------------*/
/*
	the terrain is drawn (in drawFunc), using longitude-aligned meshes,
	as follow :

		lat	^
			|
			|
   			+	1 3 5 ...(2n+1)
D_LATITUDE	|	|\|\|\	    |
			+	0 2 4 ... (2n)
			|	
			+-------+-+-------------------> lon
				D_LONGITUDE

*/
int
XCullDraw(
	float altmin, float altmax,
	int numlat, int lat_stride, float d_lat,
	int numlon, int lon_stride, float d_lon,
	void (*drawFunc)( int, int, int )
)
{
	int		j, imin, imax, jmin, jmax, imin_svg, imax_svg;
	float	latmin, latmax, lonmin, lonmax;

	xdim = (float)numlon*d_lon;
	zdim = (float)numlat*d_lat;

	numlat = (numlat-1) / lat_stride;
	numlon = (numlon-1) / lon_stride;
	d_lat *= (float) lat_stride;
	d_lon *= (float) lon_stride;

/*
	compute the 2 concave polygon resulting from the projection
	of the viewing frustum onto the planes defined by Y=altmax, Y=altmin
*/
	deleteVolume( pyramid );
	pyramid = newFrustumPyramid();
	BBoxClip( pyramid, 0., xdim, altmin, altmax, 0., zdim );

#ifdef DEV
{
	int i;
	Face *face = pyramid->firstFace;
	fprintf(stderr, "\n\n\n new pyramid:\n\n" );
	for (i=0; i<pyramid->numFace; i++, face=face->next)
	{
		int num=face->numEdge;
		Edge *edge = face->firstEdge;
		fprintf(stderr, "\n" );
		for (j=0; j<num; j++, edge = edge->next)
		{
			if (edge->inverted) fprintf( stderr, "\t\t%f %f %f,\n",
			    edge->segment->end->x,
                edge->segment->end->y,
                edge->segment->end->z
			);
			else fprintf( stderr, "\t\t%f %f %f,\n",
				edge->segment->start->x,
				edge->segment->start->y,
				edge->segment->start->z
			);
		}
	}
}
#endif
/*
	get min and max x-values of the projected viewing frustrums
	derive min and max j-indices into the elevation array
*/
	getMinMaxLat( pyramid, &latmin, &latmax );
	jmin = (int) (latmin/d_lat);
	jmax = 1 + (int) (latmax/d_lat);
	jmin = INF( SUP(jmin,0), numlat);
	jmax = INF( SUP(jmax,0), numlat);

/*
	for each x-aligned j scan-line
		get min and max z-values into the projected viewing frustrums
		derive min and max i-indices into the elevation array
		draw mesh firstVertex[j;j-1][imin;imax]
	end for
*/
#define SWAP(a,b) {int tmp=a; a=b; b=tmp;}
#define	GET_MIN_MAX_I( jval )\
	\
	if (getMinMaxLon( pyramid, (float)(jval)*d_lat, &lonmin, &lonmax ))\
	{\
		imin = (int) (lonmin/d_lon);\
		imax = 1 + (int) (lonmax/d_lon);\
		imin = INF( SUP(imin,0), numlon);\
		imax = INF( SUP(imax,0), numlon);\
	}\
	else imin = imax = -1;

	GET_MIN_MAX_I( jmin );
	imin_svg  = imin;
	imax_svg  = imax;
	for (j=jmin; j<jmax; j++)
	{
		int ifirst, ilast;

		GET_MIN_MAX_I( j+1 );

		if (imin == -1)
			ifirst = imin_svg;
		else if (imin_svg == -1)
			ifirst = imin;
		else
			ifirst = INF(imin,imin_svg);

		if (imax == -1)
			ilast = imax_svg;
		else if (imax_svg == -1)
			ilast = imax;
		else
			ilast = SUP(imax,imax_svg);

		if ((ifirst != -1) && (ilast != -1))
			(*drawFunc)( j*lat_stride, ifirst*lon_stride, ilast*lon_stride );

		imin_svg = imin;
		imax_svg = imax;
	}
}

/*---------------------------------------------------------------------------*/
/*				newFrustumPyramid
/*---------------------------------------------------------------------------*/

#define	FrontLeftBottom		0
#define	BackLeftBottom		1
#define	FrontRightBottom	2
#define	BackRightBottom		3
#define	FrontRightTop		4
#define	BackRightTop		5
#define	FrontLeftTop		6
#define	BackLeftTop			7

#define	LeftFace			0
#define	RightFace			1
#define	TopFace				2
#define	BottomFace			3
#define	FrontFace			4
#define	BackFace			5

static Volume *
newFrustumPyramid()
{
	GLint		i, vp[4];
	GLdouble	projMat[16], modelMat[16],
				x0, y0, z0, x1, y1, z1,
				*normal;
	Volume		*volume;
	Face		*Polygon[6], *face;
	Edge		*edge;
	Segment		*segment;
	Vertex		*Corner[8], *v, *v0, *v1, *v2, *v3;

/*
	Allocate the data structure for the new volume :
	------------------------------------------------
	pyramid = 8 vertex / 6 faces / 12 segments
*/
	volume = (Volume *) malloc( sizeof(Volume) );

	volume->numVertex = 8;
	for (i=0; i<8; i++) Corner[i] = (Vertex *)malloc(sizeof(Vertex));
	v = volume->firstVertex = Corner[0];
	v->last = NULL;
	for (i=1; i<8; i++)
	{
		v->next = Corner[i];
		v->next->last = v;
		v = v->next;
	}
	v->next = NULL;
	volume->lastVertex = v;

	volume->numSegment = 0;
	volume->firstSegment = volume->lastSegment = NULL;

	volume->numFace = 6;
	for (i=0; i<6; i++) Polygon[i] = newFace( 4, NULL );
	face = volume->firstFace = Polygon[0];
	face->last = NULL;
	for (i=1; i<6; i++)
	{
		face->next = Polygon[i];
		face->next->last = face;
		face = face->next;
	}
	face->next = NULL;
	volume->lastFace = face;

/*
	compute the 3D coordinates of the 8 corners of the viewing
	----------------------------------------------------------
	frustum, as resulting from the gluUnProject( current matrix )
	of the corners of the current viewport, with zval={0.,1.}
*/
	glGetDoublev( GL_PROJECTION_MATRIX, projMat );
	glGetDoublev( GL_MODELVIEW_MATRIX, modelMat );
	glGetIntegerv( GL_VIEWPORT, vp );

	x0 = (GLdouble) vp[0];
	y0 = (GLdouble) vp[1];
	x1 = (GLdouble) (vp[0]+vp[2]-1);
	y1 = (GLdouble) (vp[1]+vp[3]-1);

#define setVertex( wx, wy, wz, angle ) \
	v = Corner[angle];\
	v->in = 1;\
	gluUnProject( wx, wy, wz, modelMat, projMat, vp, &v->x, &v->y, &v->z );

	setVertex( x0, y0, 0., FrontLeftBottom );
	setVertex( x0, y0, 1., BackLeftBottom );
	setVertex( x1, y0, 0., FrontRightBottom );
	setVertex( x1, y0, 1., BackRightBottom );
	setVertex( x1, y1, 0., FrontRightTop );
	setVertex( x1, y1, 1., BackRightTop );
	setVertex( x0, y1, 0., FrontLeftTop );
	setVertex( x0, y1, 1., BackLeftTop );

/*
	now compute the 6 normals of each face of the viewing frustum
	-------------------------------------------------------------
	from the previously computed corners.
	The normals are NOT normalized, and are computed according to 
		normal = (corner0,corner1)^(corner0,corner3)

	the viewing frustum is defined by the 6 inequations :
		normal[face].(corner[face],M) >= 0
	we compute now the coefficients of these equations
*/
#define setFace( side, angle0, angle1, angle2, angle3 )\
	\
	face = Polygon[side];\
	v0 = Corner[angle0];\
	v1 = Corner[angle1];\
	v2 = Corner[angle2];\
	v3 = Corner[angle3];\
	x0 = v1->x-v0->x; y0 = v1->y-v0->y; z0 = v1->z-v0->z;\
	x1 = v3->x-v0->x; y1 = v3->y-v0->y; z1 = v3->z-v0->z;\
	normal = face->equation;\
	normal[0] = y0*z1 - y1*z0;\
	normal[1] = x1*z0 - x0*z1;\
	normal[2] = x0*y1 - x1*y0;\
	normal[3] = -(v0->x*normal[0]+v0->y*normal[1]+v0->z*normal[2]);\
	\
	edge = face->firstEdge;\
	edge->segment = newSegment( volume, v0, v1 );\
	edge->inverted = (edge->segment->start==v1);\
	edge = edge->next;\
	edge->segment = newSegment( volume, v1, v2 );\
	edge->inverted = (edge->segment->start==v2);\
	edge = edge->next;\
	edge->segment = newSegment( volume, v2, v3 );\
	edge->inverted = (edge->segment->start==v3);\
	edge = edge->next;\
	edge->segment = newSegment( volume, v3, v0 );\
	edge->inverted = (edge->segment->start==v0);

	face = volume->firstFace;
	setFace( LeftFace,
		BackLeftBottom, BackLeftTop,
		FrontLeftTop, FrontLeftBottom
	);

	face = face->next;
	setFace( RightFace,
		BackRightBottom, FrontRightBottom,
		FrontRightTop, BackRightTop
	);

	face = face->next;
	setFace( FrontFace,
		FrontLeftBottom, FrontLeftTop,
		FrontRightTop, FrontRightBottom
	);

	face = face->next;
	setFace( BottomFace,
		BackLeftBottom, FrontLeftBottom,
		FrontRightBottom, BackRightBottom
	);

	face = face->next;
	setFace( TopFace,
		BackLeftTop, BackRightTop,
		FrontRightTop, FrontLeftTop
	);

	face = face->next;
	setFace( BackFace,
		BackLeftBottom, BackRightBottom,
		BackRightTop, BackLeftTop
	);

	return volume;
}

/*---------------------------------------------------------------------------*/
/*	BBoxClip
/*---------------------------------------------------------------------------*/
static void
BBoxClip( Volume *volume,
	float xmin, float xmax,
	float ymin, float ymax,
	float zmin, float zmax )
{
	GLdouble equation[4];

#if 1
	equation[0] = 1.;		/* x-xmin >=0 */
	equation[1] = 0.;
	equation[2] = 0.;
	equation[3] = -xmin;
	ClipVolume( volume, equation );

	equation[0] = -1.;		/* -x+xmax >= 0 */
	equation[3] = xmax;
	ClipVolume( volume, equation );

	equation[0] = 0.;		/* y-ymin >= 0 */
	equation[1] = 1.;
	equation[3] = -ymin;
	ClipVolume( volume, equation );

	equation[1] = -1.;		/* -y+ymax >= 0 */
	equation[3] = ymax;
	ClipVolume( volume, equation );

	equation[1] = 0.;		/* z-zmin >= 0 */
	equation[2] = 1.;
	equation[3] = -zmin;
	ClipVolume( volume, equation );

	equation[2] = -1.;		/* -z+zmax >= 0 */
	equation[3] = zmax;
	ClipVolume( volume, equation );
#else
	equation[0] = 0.;		/* -y+ymax >= 0 */
	equation[1] = -1.;
	equation[2] = 0.;
	equation[3] = ymax;
	ClipVolume( volume, equation );
#endif
}

/*---------------------------------------------------------------------------*/
/*	ClipLine3D
/*---------------------------------------------------------------------------*/
/*
	compute the 3D intersection of the
		line: (v1,v2)
		line: eq[0]*v->x + eq[1]*v->y + eq[2]*v->z + eq[3] = 0
	(purely mathematical)
*/
static Vertex *
ClipLine3D( Vertex *v1, Vertex *v2, GLdouble *eq )
{
	GLdouble t;
	Vertex *v = (Vertex *) malloc( sizeof(Vertex) );
	v->in = 1;

	t = equate(v1,eq);
	if ( t == 0. )
	{
		v->x = v1->x;
		v->y = v1->y;
		v->z = v1->z;
	}
	else
	{
		t = t/(t-equate(v2,eq));
		v->x = v1->x + t*( v2->x - v1->x );
		v->y = v1->y + t*( v2->y - v1->y );
		v->z = v1->z + t*( v2->z - v1->z );
	}
	return v;
}

/*---------------------------------------------------------------------------*/
/*	ClipVolume
/*---------------------------------------------------------------------------*/
static void
ClipVolume( Volume *volume, GLdouble *equation )
{
	int		i, j, numf, numv, nums;
	Vertex	*v, *v0, *v1;
	Segment	*segment;
	Face	*face;
	Edge	*edge, *edge0, *edge1;

	if (volume==NULL) return;
	if ((!volume->numVertex) || (!volume->numSegment) || (!volume->numFace))
		return;

	volume->numNewSegment = 0;
	volume->firstNewSegment = NULL;
/*
	test every vertex against clipping plane
	----------------------------------------
*/
	numv = volume->numVertex;
	v = volume->firstVertex;
	for (i=0; i<numv; i++, v=v->next)
		v->in = (equate(v,equation) >= 0);
/*
	Now, rework segments
	--------------------
*/
	nums = volume->numSegment;
	segment = volume->firstSegment;
	for (i=0; i<nums; i++, segment=segment->next)
	{
		if (segment->start->in)
			segment->state = (segment->end->in) ? CULLED_IN : CLIPPED_END;
		else
			segment->state = (segment->end->in) ? CLIPPED_START : CULLED_OUT;

		if ((segment->state==CLIPPED_START) || (segment->state==CLIPPED_END))
		{
			v = ClipLine3D( segment->start, segment->end, equation);
			volume->lastVertex->next = v;
			v->next = NULL;
			v->last = volume->lastVertex;
			volume->lastVertex = v;
			volume->numVertex++;
	
			if (segment->state==CLIPPED_START) segment->start = v;
			else segment->end = v;
		}
	}

/*
	clip every face of the Volume
	-----------------------------
*/
#define	GETSTATE(edge)\
	((edge->inverted) ?\
		((edge->segment->state==CLIPPED_END) ?\
			CLIPPED_START\
			:\
			((edge->segment->state==CLIPPED_START) ?\
				CLIPPED_END\
				:\
				edge->segment->state\
			)\
		)\
		:\
		edge->segment->state\
	)

	numf = volume->numFace;
	face = volume->firstFace;
	for (i=0; i<numf; i++, face=face->next)
	{
		Edge	*edge = face->firstEdge;
		int		nume = face->numEdge;
		int		state = GETSTATE(edge);

		switch	(state)
		{
			case CULLED_IN:
			case CLIPPED_END:
				if (state!=CLIPPED_END) do
				{
					edge = edge->next;
					state = GETSTATE(edge);
				}
				while ((state!=CLIPPED_END) && (edge!=face->firstEdge));
				if ( state == CLIPPED_END )
				{
					edge0 = edge;
					edge = edge->next;
					state = GETSTATE(edge);
					while ((state!=CLIPPED_START) && (edge!=edge0))
					{
						edge = edge->next;
						state = GETSTATE(edge);
					}
					if (state != CLIPPED_START)
					{
						fprintf(stderr,
							"ERROR: ClipVolume: unable to clip face\n");
						break;
					}
	
					segment = edge0->segment;
					v0 = (edge0->inverted) ? segment->start : segment->end;
					segment = edge->segment;
					v1 = (edge->inverted) ? segment->end : segment->start;
					segment = newSegment( volume, v0, v1 );
	
					edge1 = (Edge *) malloc( sizeof(Edge) );
					edge1->segment = segment;
					edge1->inverted = (segment->start==v1);
					edge1->next = edge;
					edge1->last = edge0;
					edge = edge->last;
					while (edge != edge0)
					{
						edge = edge->last;
						free( edge->next );
						nume--;
					}
					edge1->next->last = edge1;
					edge1->last->next = edge1;
	
					face->firstEdge = edge1;
					face->numEdge = nume+1;
				}
				break;

			case CULLED_OUT:
			case CLIPPED_START:
				if (state != CLIPPED_START) do
				{
					edge = edge->next;
					state = GETSTATE(edge);
				}
				while ((state!=CLIPPED_START) && (edge!=face->firstEdge));
				if ( state != CLIPPED_START )
				{
					for (j=0; j<nume; j++)
					{
						Edge	*nextEdge = edge->next;
						free( edge );
						edge = nextEdge;
					}
					face->numEdge = 0;
					face->firstEdge = NULL;
				}
				else
				{
					edge0 = edge;
					edge = edge->next;
					state = GETSTATE(edge);
					while ((state != CLIPPED_END) && (edge!=edge0))
					{
						edge = edge->next;
						state = GETSTATE(edge);
					}
					if (state != CLIPPED_END)
					{
						fprintf(stderr,
							"ERROR: ClipVolume: unable to clip face\n");
						break;
					}
					segment = edge->segment;
					v0 = (edge->inverted) ? segment->start : segment->end;
					segment = edge0->segment;
					v1 = (edge0->inverted) ? segment->end : segment->start;
					segment = newSegment( volume, v0, v1 );

					edge1 = (Edge *) malloc( sizeof(Edge) );
					edge1->segment = segment;
					edge1->inverted = (segment->start==v1);
					edge1->next = edge0;
					edge1->last = edge;
					edge = edge->next;
					while (edge!=edge0)
					{
						edge = edge->next;
						free( edge->last );
						nume--;
					}
					edge1->next->last = edge1;
					edge1->last->next = edge1;

					face->firstEdge = edge1;
					face->numEdge = nume+1;
				}
				break;
		}
	}
/*
	build a new face, made of all the newly created segments
	--------------------------------------------------------
*/
#ifdef DEV
	fprintf( stderr, "numNewSegment = %d\n", volume->numNewSegment );
#endif
	if ((nums=volume->numNewSegment) > 0)
	{
		face = newFace( nums, equation );
		edge = face->firstEdge;
		segment = volume->firstNewSegment;
		for (i=0; i<nums; i++)
		{
			edge->segment = segment;
			segment = segment->next;
			edge = edge->next;
		}
/*
		re-order the edges, so that each segment starts with
		the end of the previous segment (of the previous edge)
*/
		edge0 = face->firstEdge;
		edge0->inverted = 0;
		for (i=0; i<nums-1; i++)
		{
			int found;
			segment = edge0->segment;
			v = (edge0->inverted) ? segment->start : segment->end;
			edge = edge0->next;
			for (found=0, j=i+1; (!found) && (j<nums); j++)
			{
				segment = edge->segment;
				if ((segment->start == v) || (segment->end == v))
				{
					found = 1;
					edge->inverted = (segment->end == v);
				}
				else edge = edge->next;
			}
			if (found)
			{
				edge->next->last = edge->last;
				edge->last->next = edge->next;
				edge->next = edge0->next;
				edge->last = edge0;
				edge->next->last = edge;
				edge->last->next = edge;
				edge0 = edge;
			}
			else
			{
				fprintf(stderr,
					"ERROR::ClipVolume:: unable to close new face\n");
				edge0 = edge0->next;
			}
		}
		face->last = volume->lastFace;
		face->next = NULL;
		volume->lastFace->next = face;
		volume->lastFace = face;
		volume->numFace++;
	}
/*
	Finally remove all the useless data :
	--------------------------------------
		vertex removed if !(vertex->in)
		segment removed if (segment->state==CULLED_OUT)
		face removed if (face->numEdge==0)
*/
#define	CLEANUP( Type, condition, firstItem, lastItem, numItem )\
	{\
		Type *ptr = firstItem;\
		int num = numItem;\
		for (i=0; i<num; i++)\
		{\
			Type *nextptr = ptr->next;\
			if (condition)\
			{\
				Type *lastptr = ptr->last;\
				if (lastptr) lastptr->next = nextptr;\
				if (nextptr) nextptr->last = lastptr;\
				if (firstItem==ptr) firstItem=nextptr;\
				if (lastItem==ptr) lastItem=lastptr;\
				free( ptr );\
				numItem--;\
			}\
			ptr = nextptr;\
		}\
	}

#if 0
	CLEANUP(
		Vertex, (!ptr->in),
		volume->firstVertex, volume->lastVertex, volume->numVertex
	);
	CLEANUP(
		Segment, (ptr->state==CULLED_OUT),
		volume->firstSegment, volume->lastSegment, volume->numSegment
	);
	CLEANUP(
		Face, (ptr->numEdge==0),
		volume->firstFace, volume->lastFace, volume->numFace
	);
#else

	{		 		
		Vertex *ptr =  		volume->firstVertex;		
		int num =  volume->numVertex 	;		
		for (i=0; i<num; i++)		{			 		
			Vertex *nextptr = ptr->next;			
			if ( (!ptr->in))			{				 		
				Vertex *lastptr = ptr->last;				
				if (lastptr) lastptr->next = nextptr;				
				if (nextptr) nextptr->last = lastptr;				
				if (volume->firstVertex==ptr)
			 		volume->firstVertex=nextptr;				
				if ( volume->lastVertex==ptr)  volume->lastVertex=lastptr;				
				free( ptr );				 
				volume->numVertex 	--;			
			}			
			ptr = nextptr;		
		}	
	};



	{		 		
		Segment *ptr =  		volume->firstSegment;		
		int num =  volume->numSegment 	;		
		for (i=0; i<num; i++)		{			 		
			Segment *nextptr = ptr->next;			
			if ( (ptr->state==1))			{				 		
				Segment *lastptr = ptr->last;				
				if (lastptr) lastptr->next = nextptr;				
				if (nextptr) nextptr->last = lastptr;				
				if (volume->firstSegment==ptr)
			 		volume->firstSegment=nextptr;				
				if ( volume->lastSegment==ptr)  volume->lastSegment=lastptr;				
				free( ptr );				 
				volume->numSegment 	--;			
			}			
			ptr = nextptr;		
		}	
	};



	{		 		
		Face *ptr =  		volume->firstFace;		
		int num =  volume->numFace 	;		
		for (i=0; i<num; i++)		{			 		
			Face *nextptr = ptr->next;			
			if ( (ptr->numEdge==0))			{				 		
				Face *lastptr = ptr->last;				
				if (lastptr) lastptr->next = nextptr;				
				if (nextptr) nextptr->last = lastptr;				
				if (volume->firstFace==ptr)
			 		volume->firstFace=nextptr;				
				if ( volume->lastFace==ptr)  volume->lastFace=lastptr;				
				free( ptr );				 
				volume->numFace 	--;			
			}			
			ptr = nextptr;		
		}	
	};
#endif
}

/*---------------------------------------------------------------------------*/
/*	deleteVolume
/*---------------------------------------------------------------------------*/
static void
deleteVolume( Volume *volume )
{
	int j,numf;
	Face *face;

	if (volume == NULL) return;

#define	DELETE( Type, numItem, firstItem )\
	{\
		int i, num = numItem;\
		Type *ptr = firstItem;\
		for (i=0; i<num; i++)\
		{\
			Type *next = ptr->next;\
			free( ptr );\
			ptr = next;\
		}\
	}

	DELETE( Vertex, volume->numVertex, volume->firstVertex );
	DELETE( Segment, volume->numSegment, volume->firstSegment );

	numf = volume->numFace;
	face = volume->firstFace;
	for (j=0; j<numf; j++)
	{
		Face *nextface = face->next;
		DELETE( Edge, face->numEdge, face->firstEdge );
		free( face );
		face = nextface;
	}

	free( volume );
}

/*---------------------------------------------------------------------------*/
/*	newSegment
/*---------------------------------------------------------------------------*/
static Segment *
newSegment( Volume *volume, Vertex *v0, Vertex *v1 )
{
	int i, num = volume->numSegment;
	Segment *segment = volume->firstSegment;

	for (i=0; i<num; i++, segment=segment->next)
	{
		if (((segment->start==v0) && (segment->end==v1)) ||
			((segment->start==v1) && (segment->end==v0)))
			return segment;
	}

	segment = (Segment *) malloc( sizeof(Segment) );
	if (!volume->numNewSegment)
		volume->firstNewSegment = segment;
	volume->numNewSegment++;

	segment->state = CULLED_IN;
	segment->start = v0;
	segment->end = v1;
	if (volume->lastSegment == NULL)
	{
		volume->firstSegment = volume->lastSegment = segment;
		segment->last = segment->next = NULL;
	}
	else
	{
		volume->lastSegment->next = segment;
		segment->last = volume->lastSegment;
		volume->lastSegment = segment;
	}
	segment->next = NULL;
	volume->numSegment++;
}

/*---------------------------------------------------------------------------*/
/*	newFace
/*---------------------------------------------------------------------------*/
static Face *
newFace( int numEdge, GLdouble *equation )
{
	int i;
	Face *face = (Face *) malloc( sizeof(Face) );
	Edge *edge;

	face->numEdge = numEdge;
	edge = face->firstEdge = (Edge *) malloc( sizeof(Edge) );
	for (i=1; i<numEdge; i++)
	{
		edge->next = (Edge *) malloc( sizeof(Edge) );
		edge->next->last = edge;
		edge = edge->next;
	}
	edge->next = face->firstEdge;	/* ring buffer */
	face->firstEdge->last = edge;

	if (equation != NULL)
		for (i=0; i<4; i++)
			face->equation[i] = equation[i];

	return face;
}

/*---------------------------------------------------------------------------*/
/*	getMinMaxLat
/*---------------------------------------------------------------------------*/
/*
	get min and max z-values of volume
*/
static int
getMinMaxLat( Volume *volume, float *latmin, float *latmax )
{
	int i, num;
	Vertex *v;
	float min, max;

	if (volume == NULL) return 0;
	if ((num=volume->numVertex) <= 0) return 0;
	if ((v=volume->firstVertex) == NULL) return 0;

	min = max = v->z;
	v = v->next;
	for (i=1; i<num; i++, v=v->next)
	{
		min = INF(min, v->z );
		max = SUP(max, v->z );
	}

	*latmin = min;
	*latmax = max;
}

/*---------------------------------------------------------------------------*/
/*	getMinMaxLon
/*---------------------------------------------------------------------------*/
/*
	for a given z=cste line, get min&max x-values of volume
*/
static int
getMinMaxLon( Volume *volume, float lat, float *lonmin, float *lonmax )
{
	int		i, num, first_found = 0;
	float	dl, lon, lonA, latA, lonB, latB, max, min;
	Segment	*segment;

	if (volume == NULL) return 0;
	if ((num=volume->numSegment) <= 0) return 0;
	if ((segment=volume->firstSegment) == NULL) return 0;

	for (i=0; i<num; i++, segment=segment->next)
	{
		lonA = segment->start->x;
		latA = segment->start->z;
		lonB = segment->end->x;
		latB = segment->end->z;
		
		if ((lat==latA) && (latA==latB))
		{
			if (first_found)
			{
				min = INF( min, INF(lonA,lonB));
				max = SUP( max, SUP(lonA,lonB));
			}
			else
			{
				min = INF(lonA,lonB);
				max = SUP(lonA,lonB);
				first_found = 1;
			}
		}
		else if (latA != latB)
		{
			dl = latB - latA;
			if (((dl>0) && (lat>=latA) && (lat<=latB)) ||
				((dl<0) && (lat<=latA) && (lat>=latB)))
			{
				lon = lonA + (lonB-lonA)*((lat-latA)/dl);
				if (first_found)
				{
					max = SUP( max, lon );
					min = INF( min, lon );
				}
				else
				{
					max = min = lon;
					first_found = 1;
				}
			}
		}
	}

	*lonmin = min;
	*lonmax = max;
	return first_found;
}
